/*	layout.c

	A simple demostration of the importance of memory layout.

	Compile-time define parameters:

	LAYOUT	set to 0-7 to select memory layout
	LOOPORD	set to 0-1 to select loop nest order
	BODY	set to loop body, e.g., A(x,y)=0;

	2022 by H. Dietz
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifndef	LAYOUT
#define	LAYOUT	0
#endif

#ifndef	LOOPORD
#define	LOOPORD	0
#endif

#ifndef	BODY
#define	BODY	A(x,y) += B(x,y);
#endif

#define	REPEATS	2
#define	X	512
#define	Y	(128*1024)

// helper macro to make a string
#define	TOSTR_(A)	#A
#define	TOSTR(A)	TOSTR_(A)

typedef	unsigned int	data_t;

typedef	struct {
	data_t	a;
	char	c;
	data_t	b;
	char	d;
} acbd_t;

typedef	struct {
	data_t	a, b;
	char	c, d;
} abcd_t;

typedef	struct {
	data_t	a, b;
} ab_t;

typedef	struct {
	char	c, d;
} cd_t;

#if	LAYOUT==0

struct {
	acbd_t	abcd[Y][X];
} data;

#define	LAYSTR	"struct {a,c,b,d} [Y][X]"
#define	A(x,y)	(data.abcd[y][x].a)
#define	B(x,y)	(data.abcd[y][x].b)
#define	C(x,y)	(data.abcd[y][x].c)
#define	D(x,y)	(data.abcd[y][x].d)

#elif	LAYOUT==1

struct {
	abcd_t	abcd[Y][X];
} data;

#define	LAYSTR	"struct {a,b,c,d} [Y][X]"
#define	A(x,y)	(data.abcd[y][x].a)
#define	B(x,y)	(data.abcd[y][x].b)
#define	C(x,y)	(data.abcd[y][x].c)
#define	D(x,y)	(data.abcd[y][x].d)

#elif	LAYOUT==2

struct {
	ab_t	ab[Y][X];
	cd_t	cd[Y][X];
} data;

#define	LAYSTR	"struct {a,b} [Y][X], struct {c,d} [Y][X]"
#define	A(x,y)	(data.ab[y][x].a)
#define	B(x,y)	(data.ab[y][x].b)
#define	C(x,y)	(data.cd[y][x].c)
#define	D(x,y)	(data.cd[y][x].d)

#elif	LAYOUT==3

struct {
	data_t	a[Y][X];
	data_t	b[Y][X];
	char	c[Y][X];
	char	d[Y][X];
} data;

#define	LAYSTR	"a[Y][X], b[Y][X], c[Y][X], d[Y][X]"
#define	A(x,y)	(data.a[y][x])
#define	B(x,y)	(data.b[y][x])
#define	C(x,y)	(data.c[y][x])
#define	D(x,y)	(data.d[y][x])

#elif	LAYOUT==4

struct {
	acbd_t	abcd[X][Y];
} data;

#define	LAYSTR	"struct {a,c,b,d} [X][Y]"
#define	A(x,y)	(data.abcd[x][y].a)
#define	B(x,y)	(data.abcd[x][y].b)
#define	C(x,y)	(data.abcd[x][y].c)
#define	D(x,y)	(data.abcd[x][y].d)

#elif	LAYOUT==5

struct {
	abcd_t	abcd[X][Y];
} data;

#define	LAYSTR	"struct {a,b,c,d} [X][Y]"
#define	A(x,y)	(data.abcd[x][y].a)
#define	B(x,y)	(data.abcd[x][y].b)
#define	C(x,y)	(data.abcd[x][y].c)
#define	D(x,y)	(data.abcd[x][y].d)

#elif	LAYOUT==6

struct {
	ab_t	ab[X][Y];
	cd_t	cd[X][Y];
} data;

#define	LAYSTR	"struct {a,b} [X][Y], struct {c,d} [X][Y]"
#define	A(x,y)	(data.ab[x][y].a)
#define	B(x,y)	(data.ab[x][y].b)
#define	C(x,y)	(data.cd[x][y].c)
#define	D(x,y)	(data.cd[x][y].d)

#else

struct {
	data_t	a[Y][X];
	data_t	b[Y][X];
	char	c[Y][X];
	char	d[Y][X];
} data;

#define	LAYSTR	"a[X][Y], b[X][Y], c[X][Y], d[X][Y]"
#define	A(x,y)	(data.a[x][y])
#define	B(x,y)	(data.b[x][y])
#define	C(x,y)	(data.c[x][y])
#define	D(x,y)	(data.d[x][y])

#endif

#if	LOOPORD==0
#define	FORXY \
	for (register int x=0; x<X; ++x) \
		for (register int y=0; y<Y; ++y)
#else
#define	FORXY \
	for (register int y=0; y<Y; ++y) \
		for (register int x=0; x<X; ++x)
#endif

int
main(int argc, char **argv)
{
	printf(
"LAYOUT=" TOSTR(LAYOUT) ", LOOPORD=" TOSTR(LOOPORD) "\n"
"X=" TOSTR(X) ", Y=" TOSTR(Y) ", sizeof(data)=%lu\n"
"for (register int r=0; r<" TOSTR(REPEATS) "; ++r)\n"
TOSTR(FORXY) " {\n"
"\t%s\n"
"}\n",
	       sizeof(data),
	       TOSTR(BODY));

	for (register int r=0; r<REPEATS; ++r)
	FORXY {
		BODY
	}

	return(0);
}
